////////////////////////////////////////////////////////////////////////
// FILE:  block.cpp
////////////////////////////////////////////////////////////////////////

#include <GL/glut.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "lib_math.h"
#include "main.h"
#include "block.h"

///////////////////////////////////////////////////
// FN:  BLOCK::initDim()
///////////////////////////////////////////////////

void BLOCK::initDim(float arg_hw, float arg_hh, float arg_hd, float arg_mass)
{
	hw = arg_hw;
	hh = arg_hh;
	hd = arg_hd;
	mass = arg_mass;
	float Ixx = mass*(hd*hd + hh*hh)/3;
	float Iyy = mass*(hh*hh + hw*hw)/3;
	float Izz = mass*(hw*hw + hd*hd)/3;
	zero(IB);
	IB[0][0] = Ixx;
	IB[0][1] = 0;
	IB[0][2] = 0;
	IB[1][0] = 0;
	IB[1][1] = Iyy;
	IB[1][2] = 0;
	IB[2][0] = 0;
	IB[2][1] = 0;
	IB[2][2] = Izz;

	const float Cvertex[][3] = {
		-hw, -hd, -hh,  // v0
		 hw, -hd, -hh,  // v1
		 hw,  hd, -hh,  // v2
		-hw,  hd, -hh,  // v3
		-hw, -hd,  hh,  // v4
		 hw, -hd,  hh,  // v5
		 hw,  hd,  hh,  // v6
		-hw,  hd,  hh,  // v7
	};
	const int CfaceV[][4] = {      //  CCW when seen from outside the block
		3,2,1,0,        // bottom
		4,5,6,7,        // top
		3,0,4,7,        // left
		1,2,6,5,        // right
		0,1,5,4,        // front
		2,3,7,6         // back
	};
	const float CfaceColor[][3] = {
		0.5, 0.5, 0.8,    // bottom -- blue
		0.5, 0.5, 0.8,    // top -- blue
		0.7, 0.4, 0.4,    // left -- red
		0.7, 0.4, 0.4,    // right -- red
		0.4, 0.7, 0.4,    // front -- green
		0.4, 0.7, 0.4     // back -- green
	};
	
	int v,n,f;
	for (v=0; v<8; v++) {
		for (n=0; n<3; n++)
			vertex[v][n] = Cvertex[v][n];
		vertex[v][3] = 1;
	}
	for (f=0; f<6; f++) {
		for (n=0; n<4; n++)
			faceV[f][n] = CfaceV[f][n];
		for (n=0; n<3; n++) 
			faceColor[f][n] = CfaceColor[f][n];
	}
}

///////////////////////////////////////////////////
// FN:  BLOCK::initState()
///////////////////////////////////////////////////

void BLOCK::initState()
{
	P[0] = 0;
	P[1] = 0;
	P[2] = 5;
	P[3] = 1;

	V[0] = 0;
	V[1] = 0;
	V[2] = 10;
	V[3] = 0;

	Q[0] = 1;
	Q[1] = 0;
	Q[2] = 0;
	Q[3] = 0;

	L[0] = 60;
	L[1] = 3;
	L[2] = 0;
	L[3] = 0;

	quat_to_R(R, Q);	    // compute R from Q
	I_inertial(I,R,IB);     // compute I in the inertial frame
	matinv(Iinv, I);        // compute inverse of I
	buildTM(TM, R, P);      // builds the full transformation matrix

	transformVertices();    // transform vertices to world coords
}

///////////////////////////////////////////////////
// FUNC:  BLOCK::transformVertices()
// DOES:  transforms vertices from local coords to world coords
///////////////////////////////////////////////////

void BLOCK::transformVertices()
{
	for (int v=0; v<8; v++)
		mvmult(vertexW[v], TM, vertex[v]);
}

///////////////////////////////////////////////////
// FN:  BLOCK::collisionImpulse(v)
///////////////////////////////////////////////////

void BLOCK::collisionImpulse(int v)
{
	VEC4 tv;
	VEC4 vrel;    // relative velocity prior to collision
	VEC4 N, rA;
	VEC4 temp1, temp2, temp3, temp4, temp5, temp6;
	float j;

	vecsub(rA, vertexW[v], P);    // compute rA, in world coords
	N[0] = 0;                     // normal of colliding surface
	N[1] = 0;
	N[2] = 1;

	// TO DO:  compute j

	  // update linear velocity:  V+ = V- + jN/mass
	vecscale(temp4, N, j/mass);
	vecadd(V, V, temp4);      

	// update angular momentum:   L+ = L- + (rA x jN)
	vecscale(temp5, N, j);
	xprod(temp6, rA, temp5);
	vecadd(L,L,temp6);
}

///////////////////////////////////////////////////
// FN:  BLOCK::collisionCheck()
///////////////////////////////////////////////////

void BLOCK::collisionCheck()
{
	extern bool Simulate;

	int count=0;
	for (int v=0; v<8; v++) {
		// TO DO:  check for collision of vertexW[v]
		// Simulate = !Simulate;
		// collisionImpulse(v);
	}
}

///////////////////////////////////////////////////
// FN:  BLOCK::sim()
///////////////////////////////////////////////////

void BLOCK::sim() {
	VEC4 force;
	VEC4 torque;
	VEC4 dV;     // time derivative of linear velocity
	VEC4 dL;     // time derivative of angular momentum
	VEC4 dQ;     // time derivative of quaternion
	int d;

	  // accumulate external forces & torques
	force[0] = force[1] = 0;
	force[2] = mass*Grav;
	torque[0] = torque[1] = torque[2] = 0;

	  // equations of motion

	// TO DO:  compute dV:  linear acceleration
	// TO DO:  compute dL:  rate of change of angular momentum

	  // integrate

	// TO DO:  update V:  velocity
	// TO DO:  update P:  position
    // TO DO:  update L:  angular momentum
	// TO DO:  recompute omega = Iinv * L
	// TO DO:  update Q:  angular orientation
	//         note: dQ = 0.5 omega * Q,  where * signifies quaternion multiplication

	// TO DO:  update R, I, Iinv, TM

	transformVertices();     // transform vertices to world coords
	collisionCheck();        // process collisions, if any
}

///////////////////////////////////////////////////
// FN:  BLOCK::draw()
///////////////////////////////////////////////////

void BLOCK::draw(bool shadow)
{
	TMAT TMt;

	glPushMatrix();
	transpose(TMt, TM);
	glMultMatrixf((const float *) TMt);  // apply rotation

	int f,n,v;

	  // draw faces
	for (f=0; f<6; f++) {
		if (!shadow)
			glColor3fv(faceColor[f]);
		glBegin(GL_POLYGON);
		for (n=0; n<4; n++) {
			v = faceV[f][n];
			glVertex3fv( vertex[v]);
		}
		glEnd();
	}
	glPopMatrix();

	if (shadow) return;

	  // draw vertices in world coords
	glColor3f(0.1,0.9,0.9);
	for (v=0; v<8; v++) {
		float *p = vertexW[v];
		glPushMatrix();
		glTranslatef(p[0], p[1], p[2]);
		glutSolidSphere(0.1, 10,10);
		glPopMatrix();
	}
}

///////////////////////////////////////////////////
// FN:  BLOCK::print()
///////////////////////////////////////////////////

void BLOCK::print()
{
	printf("%6.3f sec  P=%6.3f %6.3f %6.3f   Q=%6.3f %6.3f %6.3f %6.3f\n", Time,
		P[0],P[1],P[2], Q[0], Q[1], Q[2], Q[3]);
	printf("           V=%6.3f %6.3f %6.3f   L=%6.3f %6.3f %6.3f\n",
		V[0],V[1],V[2], L[0], L[1], L[2]);
}
