///////////////////////////////////////////////////
// FILE:      lib_math.cpp
// CONTAINS:  math routines for solving linear systems
///////////////////////////////////////////////////

#include <stdlib.h>
#include <math.h>

#include "lib_math.h"
#include "main.h"

bool Degmode = true;

///////////////////////////////////////////////
// FUNC:  matmult(c,a,b)
//        c = a*b
///////////////////////////////////////////////

void matmult(TMAT c,TMAT a,TMAT b)
{
	register int row,col,n;
	float sum;

	for (row=0; row<MSIZE; row++) {
		for (col=0; col<MSIZE; col++) {
			for (n=0,sum=0; n<MSIZE; n++) {
				sum += a[row][n] * b[n][col];
			}
			c[row][col] = sum;
		}
	}
}

///////////////////////////////////////////////
// FUNC:  matcopy(b,a)
//        b = a
///////////////////////////////////////////////

void matcopy(TMAT b,TMAT a)
{
	register int row,col;

	for (row=0; row<MSIZE; row++)
		for (col=0; col<MSIZE; col++)
			b[row][col] = a[row][col];
}

///////////////////////////////////////////////
// FUNC:  matinv(minv, m)
//        inverts a 3x3 matrix
///////////////////////////////////////////////

void matinv(TMAT minv, TMAT m)
{
	float a = m[0][0];
	float b = m[0][1];
	float c = m[0][2];
	float d = m[1][0];
	float e = m[1][1];
	float f = m[1][2];
	float g = m[2][0];
	float h = m[2][1];
	float i = m[2][2];

	float D = a*(e*i-f*h) - b*(d*i-g*f) + c*(d*h-e*g);
	float Di = 1.0f/D;
	iden(minv);
	minv[0][0] = Di*(e*i-f*h);
	minv[0][1] = Di*(c*h-i*b);
	minv[0][2] = Di*(b*f-e*c);
	minv[1][0] = Di*(f*g-i*d);
	minv[1][1] = Di*(a*i-g*c);
	minv[1][2] = Di*(c*d-a*f);
	minv[2][0] = Di*(d*h-e*g);
	minv[2][1] = Di*(b*g-h*a);
	minv[2][2] = Di*(a*e-b*d);
	
	return;

	  // double check the result
	bool bad=false;
	int row,col;
	TMAT prod;
	zero(prod);
	for (row=0; row<3; row++) {
		for (col=0; col<3; col++) {
			float sum=0;
			for (int n=0; n<3; n++)
				sum += m[row][n]*minv[n][col];
			prod[row][col] = sum;
			if (row==col) {
				float diff = fabs(sum-1.0f);
				if (diff>0.01) {
					bad = true;
				}
			} else if (fabs(sum)>0.01f) {
				bad = true;
			}
		}
	}
	if (bad) {
		printf("bad matrix inverse\n");
		prmat(prod);
	} else {
		printf("good matrix inverse\n");
	}
}

////////////////////////////////////////////////////////////
// FUNC:  iden()
// DOES:  creates 4x4 identity matrix
////////////////////////////////////////////////////////////

void iden(TMAT mat)
{
	register int row,col;

	for (row=0; row<MSIZE; row++)
		for (col=0; col<MSIZE; col++)
			mat[row][col] = (row==col);
}

////////////////////////////////////////////////////////////
// FUNC:  zero()
// DOES:  creates 4x4 zero matrix
////////////////////////////////////////////////////////////

void zero(TMAT mat)
{
	register int row,col;

	for (row=0; row<MSIZE; row++)
		for (col=0; col<MSIZE; col++)
			mat[row][col] = 0;
}

////////////////////////////////////////////////////////////
// FUNC:  transpose()
// DOES:  tranposes a matrix
////////////////////////////////////////////////////////////

void transpose(TMAT tmat, TMAT mat)
{
	register int row,col;
	
	for (row=0; row<MSIZE; row++)
		for (col=0; col<MSIZE; col++)
			tmat[row][col] = mat[col][row];
}

////////////////////////////////////////////////////////////
// FUNC:  mvmult()
// DOES:  matrix-vector multiply:  ans = mat * vec
////////////////////////////////////////////////////////////

void mvmult(VEC4 ans,TMAT mat,VEC4 vec)
{
	register int row,n;
	float sum;

	for (row=0; row<MSIZE; row++) {
		sum = 0;
		for (n=0; n<MSIZE; n++)
			sum += vec[n] * mat[row][n];
		ans[row] = sum;
	}
}

////////////////////////////////////////////////////////////
// FUNC:  prmat(m)
// DOES:  prints 4x4 matrix m
////////////////////////////////////////////////////////////

void prmat(TMAT mat)
{
	for (int row=0; row<MSIZE; row++)
		prvec(mat[row]);
	printf("\n");
}

////////////////////////////////////////////////////////////
// FUNC:  prvec(v)
// DOES:  prints 4-vector
////////////////////////////////////////////////////////////

void prvec(VEC4 vec)
{
	for (int n=0; n<MSIZE; n++)
		printf("%9.3f ",vec[n]);
	printf("\n");
}

//////////////////////////////////////////////////
// PROC:  vecsub(ans, vec1, vec2)
//////////////////////////////////////////////////

void vecsub(VEC4 ans,VEC4 vec1,VEC4 vec2)
{
	int n;

	for (n=0; n<MSIZE; n++)
		ans[n] = vec1[n] - vec2[n];
}

//////////////////////////////////////////////////
// PROC:  vecadd(ans, vec1, vec2)
//////////////////////////////////////////////////

void vecadd(VEC4 ans,VEC4 vec1,VEC4 vec2)
{
	for (int n=0; n<MSIZE; n++)
		ans[n] = vec1[n] + vec2[n];
}

//////////////////////////////////////////////////
// PROC:  vecscale(vec, sf)
//////////////////////////////////////////////////

void vecscale(VEC4 ans, VEC4 vec, float sf)
{
	for (int n=0; n<MSIZE; n++)
		ans[n] = vec[n] * sf;
}

//////////////////////////////////////////////////
// PROC:  xprod(norm, vec1, vec2)
// DOES:  computes cross product:  norm = vec1 x vec2
//////////////////////////////////////////////////

void xprod(VEC4 norm,VEC4 vec1,VEC4 vec2)
{
	norm[0] = vec1[1]*vec2[2] - vec2[1]*vec1[2];
	norm[1] = vec2[0]*vec1[2] - vec1[0]*vec2[2];
	norm[2] = vec1[0]*vec2[1] - vec2[0]*vec1[1];
}

//////////////////////////////////////////////////
// PROC:  vecmag(vec)
// DOES:  computes magnitude of a 3-vector
//////////////////////////////////////////////////

float vecmag(VEC4 vec)
{
	float sum=0;

	for (int n=0; n<3; n++)
		sum += vec[n]*vec[n];
	return(sqrt(sum));
}

////////////////////////////////////////////////////////////
// FUNC:  dotprod(result, vec1, vec2)
// DOES:  dot-product of vectors:  result = vec1 dot vec2
////////////////////////////////////////////////////////////

void dotprod(float *result,VEC4 vec1,VEC4 vec2)
{
	float sum=0;

	for (int n=0; n<3; n++)
		sum += vec1[n]*vec2[n];
	*result = sum;
}

////////////////////////////////////////////////////////////
// FUNC:  univec(VEC4 vec)
// DOES:  normalized a 3-vector
////////////////////////////////////////////////////////////

void univec(VEC4 vec)
{
	float mag;

	mag = vecmag(vec);
	vec[0] /= mag;
	vec[1] /= mag;
	vec[2] /= mag;
}

////////////////////////////////////////////////////////////
// FUNC:  negvec(VEC4 ans, vec)
// DOES:  computes   ans = -vec
////////////////////////////////////////////////////////////

void negvec(VEC4 ansvec,VEC4 vec)
{
	ansvec[0] = -vec[0];
	ansvec[1] = -vec[1];
	ansvec[2] = -vec[2];
	ansvec[3] = -vec[3];
}

////////////////////////////////////////////////////////////
// FUNC:  vecrot(tm, vec, angle)
// DOES:  builds a 4x4 transformation matrix tm = Rot(vec, angle)
////////////////////////////////////////////////////////////

void vecrot(TMAT tm,VEC4 vec,float angle)
{
	float c,s,v;
	register float kx,ky,kz;

	kx = vec[0]; ky = vec[1]; kz = vec[2];
	c = cos(angle);
	s = sin(angle);
	v = 1-c;

	tm[0][0] = kx*kx*v + c;
	tm[0][1] = ky*kx*v - kz*s;
	tm[0][2] = kz*kx*v + ky*s;
	tm[0][3] = 0;

	tm[1][0] = kx*ky*v + kz*s;
	tm[1][1] = ky*ky*v +  c;
	tm[1][2] = kz*ky*v - kx*s;
	tm[1][3] = 0;

	tm[2][0] = kx*kz*v - ky*s;
	tm[2][1] = ky*kz*v + kx*s;
	tm[2][2] = kz*kz*v + c;
	tm[2][3] = 0;

	tm[3][0] = 0;
	tm[3][1] = 0;
	tm[3][2] = 0;
	tm[3][3] = 1;
}

///////////////////////////////////////////////////
// FN:  quatMult(q,q1,q2)
//      multiplies two quaternions
///////////////////////////////////////////////////

void quatMult(VEC4 q, VEC4 q1, VEC4 q2)
{
	float w1 = q1[0];
	float x1 = q1[1];
	float y1 = q1[2];
	float z1 = q1[3];

	float w2 = q2[0];
	float x2 = q2[1];
	float y2 = q2[2];
	float z2 = q2[3];

	float w = (w1*w2 - x1*x2 - y1*y2 - z1*z2);
	float x = (w1*x2 + x1*w2 + y1*z2 - z1*y2);
	float y = (w1*y2 - x1*z2 + y1*w2 + z1*x2);
	float z = (w1*z2 + x1*y2 - y1*x2 + z1*w2);

	q[0] = w;
	q[1] = x;
	q[2] = y;
	q[3] = z;
}

///////////////////////////////////////////////////
// FN:  quatNorm()
//      normalizes the quaternion
///////////////////////////////////////////////////

void quatNorm(VEC4 q)
{
	float sf = 1.0/sqrt(q[0]*q[0] + q[1]*q[1] + q[2]*q[2] + q[3]*q[3]);
	q[0] *= sf;
	q[1] *= sf;
	q[2] *= sf;
	q[3] *= sf;
}

///////////////////////////////////////////////////
// FN:  quat_to_R()
///////////////////////////////////////////////////

void quat_to_R(TMAT rmat, VEC4 quat)
{
	float w = quat[0];
	float x = quat[1];
	float y = quat[2];
	float z = quat[3];
	zero(rmat);
	rmat[0][0] = 1 - 2*y*y - 2*z*z;
	rmat[0][1] = 2*x*y - 2*w*z;
	rmat[0][2] = 2*x*z + 2*w*y;
	rmat[1][0] = 2*x*y + 2*w*z;
	rmat[1][1] = 1 - 2*x*x - 2*z*z;
	rmat[1][2] = 2*y*z - 2*w*x;
	rmat[2][0] = 2*x*z - 2*w*y;
	rmat[2][1] = 2*y*z + 2*w*x;
	rmat[2][2] = 1 - 2*x*x - 2*y*y;
}


