////////////////////////////////////////////////////
// FILE:   main.cpp
////////////////////////////////////////////////////

#include <GL/glut.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "lib_math.h"
#include "main.h"
#include "block.h"
#include "draw.h"

static const float WINSIZE = 512;
bool Simulate;

BLOCK Block;

void quat_to_R(TMAT rmat, VEC4 quat);      // compute R from Q
void I_inertial(TMAT i, TMAT R, TMAT IB);  // compute I in the inertial frame

double Dt;          // time step
double Time;        // absolute time
double Grav = -9.8;
double Eps = 0.5;   // coeff of restitution

const double eye[3] = { -9,-20,10};  // eye position
const double cen[3] = {0,0,0};       // center or lookat point
const double vup[3] = {0,0,1};       // view up vector
int TestFlag;

///////////////////////////////////////////////////
// FN:  init()
///////////////////////////////////////////////////

void init()
{
      // initialize matrix stacks
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
	gluPerspective(60,1,1,50);

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
	gluLookAt(eye[0], eye[1], eye[2], 
		      cen[0], cen[1], cen[2],
			  vup[0], vup[1], vup[2]);
	glViewport( 0, 0, WINSIZE, WINSIZE );
    glEnable( GL_DEPTH_TEST );
	glDepthFunc(GL_LEQUAL);
	Block.initDim(4,1,2,10);    // width, height, depth, mass
	Block.initState();
	Block.mass = 1;
	Dt = 0.01;
	Time = 0;
	Simulate = false;
	TestFlag = 10;
}

///////////////////////////////////////////////////
// FN:  buildTM()
///////////////////////////////////////////////////

void buildTM(TMAT tm, TMAT rotMat, VEC4 translate)
{
	int r,c;

	iden(tm);          // initialize to identity matrix
	for (r=0; r<3; r++) 
		for (c=0; c<3; c++)
			tm[r][c] = rotMat[r][c];
	for (r=0; r<3; r++)
		tm[r][3] = translate[r];
}

///////////////////////////////////////////////////
// FN:  I_inertial()
//      converts inertia tensor from local frame to the inertial frame
///////////////////////////////////////////////////

void I_inertial(TMAT i, TMAT R, TMAT ib)
{
	TMAT Rt;
	TMAT temp;

	  // compute I = R IB R^T
	transpose(Rt, R);
	matmult(temp, ib, Rt);
	matmult(i,R,temp);
}

///////////////////////////////////////////////////
// FN:  simStep()
///////////////////////////////////////////////////

void simStep()
{
	if (!Simulate) return;

	Block.sim();
	Time += Dt;
	glutPostRedisplay();
}

/////////////////////////////////////////////////////
//    PROC: keyCB()
//    DOES: glut keyboard call-back
/////////////////////////////////////////////////////

void keyCB(unsigned char key, int x, int y)
{
	switch(key) {
	case 'q':
		exit(0);
		break;
	case ' ':
		Simulate = !Simulate;
		break;
	case 'r':
		init();
		break;
	case '>':
		Eps += 0.1;
		if (Eps>1.1)
			Eps = 1.1;
		printf("Eps=%6.3f\n",Eps);
		break;
	case '<':
		Eps -= 0.1;
		if (Eps<0)
			Eps = 0.0;
		printf("Eps=%6.3f\n",Eps);
		break;
	default:
		printf("unrecognized key\n");
		break;
  }
  glutPostRedisplay();
}

///////////////////////////////////////////////////
// FN:  main()
///////////////////////////////////////////////////

int main(int argc, char **argv)
{
    // create window and rendering context
    glutInit( &argc, argv );
    glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );
    glutInitWindowSize( WINSIZE, WINSIZE );
    glutCreateWindow( "basicOpenGL" );
    glutDisplayFunc( displayCB );
	glutKeyboardFunc(keyCB);
	glutIdleFunc( simStep );

    init();

    glutMainLoop();
    return 0;           // never reached
}
